"""Route validation functions."""

from __future__ import annotations

import inspect
import struct
import typing

if typing.TYPE_CHECKING:
    from fastcc.utilities.type_aliases import Routable
    from fastcc.utilities.type_definitions import Packet

from google.protobuf.message import Message


def get_packet_parameter(routable: Routable) -> inspect.Parameter | None:
    """Get the packet parameter of a routable.

    Parameters
    ----------
    routable
        Routable to get the packet parameter from.

    Returns
    -------
    inspect.Parameter
        Packet parameter.
    None
        If the routable has no packet parameter.
    """
    signature = inspect.signature(routable, eval_str=True)
    parameters = list(signature.parameters.values())

    try:
        p0 = parameters.pop(0)
    except IndexError:
        p0 = None

    if p0 is None:
        return None

    if p0.kind != inspect.Parameter.POSITIONAL_OR_KEYWORD:
        return None

    return p0


def bytes_to_packet(payload: bytes, packet_type: type[Packet]) -> Packet:
    """Convert bytes to a packet.

    Parameters
    ----------
    payload
        Payload to convert.
    packet_type
        Packet type to convert to.

    Returns
    -------
    Packet
        Converted packet.
    """
    if issubclass(packet_type, bytes):
        return payload

    if issubclass(packet_type, str):
        return payload.decode()

    if issubclass(packet_type, int):
        return int.from_bytes(payload)

    if issubclass(packet_type, float):
        return struct.unpack("f", payload)[0]  # type: ignore [no-any-return]

    if issubclass(packet_type, Message):
        message = packet_type()
        message.ParseFromString(payload)
        return message

    details = f"packet type {packet_type} is not supported"
    raise NotImplementedError(details)



"""Type aliases for improved type safety and code clarity."""

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from fastcc.utilities.type_definitions import Packet

    Routable = Callable[..., Awaitable[Packet | None]]



"""Route validation functions."""

from __future__ import annotations

import inspect
import logging
import types
import typing

if typing.TYPE_CHECKING:
    from fastcc.utilities.type_aliases import Routable

from fastcc.utilities.type_definitions import Packet

_logger = logging.getLogger(__name__)


def validate_route(routable: Routable) -> None:
    """Check if a routable is valid.

    Parameters
    ----------
    routable
        Routable to validate.
    """
    signature = inspect.signature(routable, eval_str=True)
    parameters = list(signature.parameters.values())

    try:
        p0 = parameters.pop(0)
    except IndexError:
        p0 = None

    if p0 is not None:
        validate_route_first_parameter(p0)

    validate_keyword_only_parameters(parameters)
    validate_return_type(signature.return_annotation)


def validate_route_first_parameter(p0: inspect.Parameter) -> None:
    """Check if the first parameter of a route is valid.

    Parameters
    ----------
    p0
        First parameter of a route.

    Raises
    ------
    ValueError
        If the first parameter is invalid.
    """
    allowed_p0_kinds = {
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
        inspect.Parameter.KEYWORD_ONLY,
    }
    if p0.kind not in allowed_p0_kinds:
        details = (
            f"first parameter must be positional (packet) "
            f"or keyword-only (injector), not {p0.kind!r}"
        )
        _logger.error(details)
        raise ValueError(details)

    if p0.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
        # p0 is packet parameter => verify type
        if p0.annotation is inspect.Parameter.empty:
            details = "packet parameter must have a type annotation"
            _logger.error(details)
            raise ValueError(details)

        if type(p0.annotation) is types.UnionType:
            details = (
                f"packet parameter with type {types.UnionType} is "
                f"not supported yet"
            )
            _logger.error(details)
            raise NotImplementedError(details)

        try:
            success = issubclass(p0.annotation, Packet)
        except TypeError:
            success = False

        if not success:
            details = (
                f"packet parameter must be of type {Packet} not {p0.annotation}"
            )
            _logger.error(details)
            raise ValueError(details)


def validate_keyword_only_parameters(
    parameters: list[inspect.Parameter],
) -> None:
    """Check if all parameters are keyword-only.

    Parameters
    ----------
    parameters
        Parameters to validate.

    Raises
    ------
    ValueError
        If a parameter is not keyword-only.
    """
    for p in parameters:
        if p.kind != inspect.Parameter.KEYWORD_ONLY:
            details = f"parameter '{p}' must be keyword-only, not {p.kind!r}"
            _logger.error(details)
            raise ValueError(details)


def validate_return_type(return_annotation: typing.Any) -> None:  # noqa: ANN401
    """Check if the return type is valid.

    Parameters
    ----------
    return_annotation
        Return type to validate.

    Raises
    ------
    ValueError
        If the return type is invalid.
    """
    if return_annotation is inspect.Parameter.empty:
        details = "return value must have a type annotation"
        _logger.error(details)
        raise ValueError(details)

    if return_annotation is None:
        return

    if type(return_annotation) is types.UnionType:
        details = (
            f"return value with type {types.UnionType} is not "
            f"supported yet"
        )
        _logger.error(details)
        raise NotImplementedError(details)

    try:
        success = issubclass(return_annotation, Packet)
    except TypeError:
        success = False

    if not success:
        details = (
            f"return value must be of type {Packet} not {return_annotation}"
        )
        _logger.error(details)
        raise ValueError(details)



"""
Framework for component communication.

FastCC is a lightweight, efficient and developer-friendly framework for
component communication. By leveraging MQTT [1]_ for messaging and
Protocol Buffers [2]_ for data serialization, this framework ensures
fast, reliable, and bandwidth-efficient communication. Its simplicity
in setup and development makes it ideal for both small-scale and
enterprise-level applications.

References
----------
.. [1] https://mqtt.org
.. [2] https://protobuf.dev/
"""

from __future__ import annotations

import inspect
import logging
import typing

if typing.TYPE_CHECKING:
    from collections.abc import Callable

    from fastcc.utilities.type_aliases import Routable

import aiomqtt
from google.protobuf.message import Message
from paho.mqtt.packettypes import PacketTypes
from paho.mqtt.properties import Properties

from fastcc.utilities import interpretation
from fastcc.utilities.mqtt import QoS
from fastcc.utilities.validation import validate_route

__all__ = ["CCRouter", "FastCC"]

_logger = logging.getLogger(__name__)


class CCRouter:
    """Router for communication endpoints."""

    def __init__(self) -> None:
        self._routes: dict[str, dict[int, list[Routable]]] = {}

    @property
    def routes(self) -> dict[str, dict[int, list[Routable]]]:
        """Return all registered routes.

        Returns
        -------
        dict[str, dict[int, list[Routable]]]
            All registered routes.
        """
        return self._routes

    def route(
        self,
        topic: str,
        *,
        qos: QoS = QoS.AT_MOST_ONCE,
    ) -> Callable[[Routable], Routable]:
        """Register a (MQTT-) route.

        Parameters
        ----------
        topic
            MQTT topic to subscribe and assign the function to.
        qos
            Quality of Service level for the subscription.

        Returns
        -------
        Callable[[Routable], Routable]
            Decorated callable.
        """

        def decorator(routable: Routable) -> Routable:
            validate_route(routable)

            if topic not in self._routes:
                self._routes[topic] = {}

            if qos not in self._routes[topic]:
                self._routes[topic][qos] = []

            self._routes[topic][qos].append(routable)

            return routable

        return decorator

    def add_router(self, router: CCRouter) -> None:
        """Add another router to this router.

        Parameters
        ----------
        router
            Router to add.
        """
        self._routes.update(router.routes)


class FastCC:
    """Application class of FastCC.

    Parameters
    ----------
    client
        MQTT client instance.

    Raises
    ------
    ValueError
        If `client` does not support MQTT version 5.0.
    """

    def __init__(self, client: aiomqtt.Client) -> None:
        if client._client.protocol != aiomqtt.ProtocolVersion.V5.value:  # noqa: SLF001
            details = "client must support MQTT version 5.0"
            _logger.error(details)
            raise ValueError(details)

        self._client = client
        self._router = CCRouter()
        self._injectors: dict[str, typing.Any] = {}

    async def run(self) -> None:
        """Start the application.

        Raises
        ------
        ValueError
            If the initialized MQTT client is not connected.
        """
        if not self._client._connected.done():  # noqa: SLF001
            details = "client is not connected"
            _logger.error(details)
            raise ValueError(details)

        for topic, data in self._router.routes.items():
            for qos in data:
                _logger.info("Subscribing to topic %s (QoS %d)", topic, qos)
                await self._client.subscribe(topic, qos)

        await self.__listen()

    def add_router(self, router: CCRouter) -> None:
        """Add a router to the app.

        Parameters
        ----------
        router
            Router to add.
        """
        self._router.add_router(router)

    def add_injector(self, **kwargs: typing.Any) -> None:  # noqa: ANN401
        """Add injector variables to the app.

        Injector variables are passed to the routables as keyword
        arguments if they are present (by name!).
        """
        self._injectors.update(kwargs)

    async def __listen(self) -> None:
        async for message in self._client.messages:
            await self.__handle(message)

    async def __handle(self, message: aiomqtt.Message) -> None:  # noqa: C901
        if not isinstance(message.payload, bytes):
            details = (
                f"message type was expected to be bytes but was "
                f"{type(message.payload)}"
            )
            _logger.error(details)
            raise TypeError(details)

        if (routings := self._router.routes.get(message.topic.value)) is None:
            return

        if (callbacks := routings.get(QoS(message.qos))) is None:
            return

        for callback in callbacks:
            signature = inspect.signature(callback, eval_str=True)

            kwargs = {
                key: value
                for key, value in self._injectors.items()
                if key in signature.parameters
            }

            packet_parameter = interpretation.get_packet_parameter(callback)
            if packet_parameter is not None:
                packet = interpretation.bytes_to_packet(
                    message.payload,
                    packet_parameter.annotation,
                )
                kwargs[packet_parameter.name] = packet

            response = None
            properties = Properties(PacketTypes.PUBLISH)  # type: ignore [no-untyped-call]
            try:
                response = await callback(**kwargs)
            except Exception as error:
                _logger.exception("error while handling message")
                properties.UserProperty = [("error", error.__class__.__name__)]

            response_topic = getattr(message.properties, "ResponseTopic", None)

            if response_topic is not None:
                if isinstance(response, Message):
                    response = response.SerializeToString()

                await self._client.publish(
                    response_topic,
                    response,
                    qos=message.qos,
                    properties=properties,
                )



"""Utilities related to MQTT."""

from __future__ import annotations

import enum


class QoS(enum.IntEnum):
    """Quality of Service levels [1]_.

    References
    ----------
    .. [1] https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html#_Toc3901234
    """

    #: The message is delivered at most once, or it is not delivered at all.
    AT_MOST_ONCE = 0

    #: The message is always delivered at least once.
    AT_LEAST_ONCE = 1

    #: The message is always delivered exactly once.
    EXACTLY_ONCE = 2
